package de.tud.kom.p2psim.impl.loadgen;

import java.util.ArrayList;
import java.util.List;

import de.tud.kom.p2psim.impl.loadgen.config.Request;
import de.tud.kom.p2psim.impl.loadgen.config.Scenario;
import de.tud.kom.p2psim.impl.loadgen.nodeselection.AbstractNodeSelection;
import de.tud.kom.p2psim.impl.loadgen.nodeselection.AbstractNodeSelection.IllegalSelectionException;
import de.tudarmstadt.maki.simonstrator.api.Event;
import de.tudarmstadt.maki.simonstrator.api.EventHandler;
import de.tudarmstadt.maki.simonstrator.api.Host;
import de.tudarmstadt.maki.simonstrator.api.Oracle;
import de.tudarmstadt.maki.simonstrator.api.Time;
import de.tudarmstadt.maki.simonstrator.api.common.graph.INodeID;
import de.tudarmstadt.maki.simonstrator.api.util.XMLConfigurableConstructor;

public class LoadGenerator implements EventHandler {
	// Event Types
	private static final int INCOMING_REQUEST = 35185;

	private static final int SIMULATION_STARTED = 35186;

	// Storage
	private Scenario scenario;

	private long sampleRate;

	private AbstractNodeSelection nodeSelection;

	/**
	 * Constructor
	 * 
	 * @param sampleRate
	 *            the rate at which the load generator should emit events
	 */
	@XMLConfigurableConstructor({ "sampleRate" })
	public LoadGenerator(String sampleRate) {
		this.sampleRate = Time.parseTime(sampleRate);
		Event.scheduleImmediately(this, null, SIMULATION_STARTED);
	}

	/**
	 * Set the scenario for the load gen
	 * 
	 * @param scenario
	 *            An Object that contains Requests, Sources, Targets &
	 *            Distribution
	 */
	public void setScenario(Scenario scenario) {
		this.scenario = scenario;
		// Schedule each request in scenario the scenario at sample time
		for (Request req : this.scenario.getRequests()) {
			Event.scheduleWithDelay(sampleRate, this, req, INCOMING_REQUEST);
		}
	}

	/**
	 * Set the node selection algorithm for the scenario.
	 * 
	 * @param nodeSelection
	 */
	public void setNodeSelection(AbstractNodeSelection nodeSelection) {
		this.nodeSelection = nodeSelection;
		// For some node selection algorithms it is necessary that it knows all
		// nodes that are present in the network
		this.nodeSelection.setAllNodeIDs(Filter.getAllNodeIDs());
	}

	@Override
	public void eventOccurred(Object content, int type) {
		switch (type) {
		case SIMULATION_STARTED:
			// Simulation started
			System.out.println("Simulation started!");
			break;
		case INCOMING_REQUEST:
			// Received a request
			System.out.println("Received REQUEST_TYPE");
			if (content == null) {
				System.out
						.println("Received a Request without a valid object.");
				break;
			}

			handleRequest((Request) content);
			break;
		default:
			System.out.println("Unknown Event received: " + type);
			break;
		}
	}

	/**
	 * Invokes a request at source nodes, according to distribution and node
	 * selection.
	 * 
	 * @param request
	 *            Request
	 */
	private void handleRequest(Request request) {

		// Schedule next request directly after next sampling time
		Event.scheduleWithDelay(sampleRate, this, request, INCOMING_REQUEST);

		System.out.println("Received Request");
		System.out.println(request.toString());

		// Get all active nodes in the network
		ArrayList<INodeID> activeNodes = Filter.getActiveNodes();
		System.out.println("Active Nodes: " + activeNodes.size());

		if (activeNodes.size() == 0) {
			System.out.println("No active nodes in Network");
			return;
		}

		// Check each active node if it matches all conditions
		ArrayList<INodeID> filteredNodes = new ArrayList<>();
		for (INodeID nodeId : activeNodes) {
			if (Utils.matchNodeAgainstSetOfConditions(nodeId,
					request.getSources())) {
				filteredNodes.add(nodeId);
			}
		}
		System.out.println(
				"Nodes that match conditions: " + filteredNodes.size());

		if (filteredNodes.size() == 0) {
			System.out.println("No nodes match all conditions");
			return;
		}

		// Choose amount of nodes from distribution
		int nodeCount = request.getDistribution()
				.getCount(Time.getCurrentTime(), filteredNodes.size());

		System.out.println("Node count from Distribution: " + nodeCount);
		if (nodeCount == 0) {
			System.out.println("No nodes selected from Distribution");
			return;
		}

		// Node Selection and invoking
		try {
			List<INodeID> selectedNodes = this.nodeSelection
					.selectNodes(nodeCount, filteredNodes);

			for (INodeID nodeID : selectedNodes) {
				System.out.println("Invoke Request on " + nodeID.toString()
						+ ": " + request.getRequestType().toString());

				Host currentHost = Oracle.getHostByID(nodeID);
				if (currentHost instanceof RequestInvokable)
					((RequestInvokable) currentHost).invokeRequest(request);
				else
					System.err.println(
							"Host with ID " + nodeID + " doesn't implement "
									+ RequestInvokable.class.getName());
			}

		} catch (IllegalSelectionException e) {
			e.printStackTrace();
		}
	}
}
