package de.tud.kom.p2psim.impl.resourcedistribution;

import de.tud.kom.p2psim.api.analyzer.ResourceAnalyzer;
import de.tud.kom.p2psim.api.common.SimHost;
import de.tud.kom.p2psim.api.energy.EnergyComponent;
import de.tud.kom.p2psim.api.energy.EnergyEventListener;
import de.tud.kom.p2psim.api.energy.EnergyState;
import de.tud.kom.p2psim.api.network.SimNetInterface;
import de.tud.kom.p2psim.api.resourcedistribution.ResourceChangeListener;
import de.tud.kom.p2psim.api.resourcedistribution.ResourceModel;
import de.tudarmstadt.maki.simonstrator.api.Monitor;
import de.tudarmstadt.maki.simonstrator.api.Time;
import de.tudarmstadt.maki.simonstrator.api.component.ComponentNotAvailableException;
import de.tudarmstadt.maki.simonstrator.api.component.core.MonitorComponent;
import de.tudarmstadt.maki.simonstrator.api.component.resourcedistribution.ResourceConsumer;
import de.tudarmstadt.maki.simonstrator.api.component.resourcedistribution.ResourceConsumptionState;
import de.tudarmstadt.maki.simonstrator.api.component.sensor.resource.ResourceSensor;

import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

public class DefaultResourceModel implements ResourceModel, ResourceSensor, EnergyEventListener {
    private SimHost host;
    private Map<Short, Resource> resourceInformation;
    private Set<ResourceChangeListener> resourceChangeListeners;

    private Map<EnergyComponent, Short> energyComponents = new HashMap<>();
    private ResourceConsumer consumer;

    public DefaultResourceModel(SimHost host) {
        this.host = host;
        resourceInformation = new HashMap<>();
        resourceChangeListeners = new HashSet<>();
    }

    @Override
    public void initialize() {
        try {
            this.consumer = host.getComponent(ResourceConsumer.class);
        } catch (ComponentNotAvailableException e) {
            throw new RuntimeException("Resource model requires a resource consumer component on host", e);
        }

        try {
            ResourceAnalyzer resourceAnalyzer = Monitor.get(ResourceAnalyzer.class);
            resourceAnalyzer.initialize(consumer, this);
        } catch (MonitorComponent.AnalyzerNotAvailableException e) {
            // no analyzer present
        }
    }

    @Override
    public ResourceSensor getResourceInformation() {
        return this;
    }

    @Override
    public boolean isVital(short resourceType) {
        if (!resourceInformation.containsKey(resourceType)) {
            throw new IllegalArgumentException("Unknown resource type " + resourceType);
        }
        return resourceInformation.get(resourceType).isVital;
    }

    @Override
    public void registerResourceType(short resourceTypeId, String name, double maxValue, double initialValue, boolean
            vital) {
        if (resourceInformation.containsKey(resourceTypeId)) return;

        Resource resource = new Resource();
        resource.currentValue = initialValue;
        resource.maxValue = maxValue;
        resource.id = resourceTypeId;
        resource.name = name;
        resource.isVital = vital;

        resourceInformation.put(resourceTypeId, resource);
    }

    @Override
    public void addResourceChangeListener(ResourceChangeListener listener) {
        resourceChangeListeners.add(listener);
    }

    @Override
    public boolean removeResourceChangeListener(ResourceChangeListener listener) {
        return resourceChangeListeners.remove(listener);
    }

    @Override
    public void registerComponent(EnergyComponent comp, short resourceTypeId) {
        if (!resourceInformation.containsKey(resourceTypeId)) {
            throw new IllegalArgumentException("Unknown resource with ID " + resourceTypeId);
        }
        comp.setEnergyEventListener(this);
        energyComponents.put(comp, resourceTypeId);
    }

    @Override
    public void fillResourceUnits(short resourceTypeId, double resourceUnits) {
        if (!resourceInformation.containsKey(resourceTypeId)) {
            throw new IllegalArgumentException("Unknown resource with ID " + resourceTypeId);
        }
        if (resourceUnits < 0) {
            throw new IllegalArgumentException(String.format("Number of resource units must be positive, %f given",
                    resourceUnits));
        }

        resourceInformation.get(resourceTypeId).increase(resourceUnits);

    }

    @Override
    public void _fillResourceUnits(short resourceTypeId, double resourceUnits) {
        fillResourceUnits(resourceTypeId, resourceUnits);
    }
    
    @Override
	public void setResourcePercentage(short resourceTypeId, double resourcePercentage) {
    	 if (!resourceInformation.containsKey(resourceTypeId)) {
             throw new IllegalArgumentException("Unknown resource with ID " + resourceTypeId);
         }
         if (resourcePercentage < 0 || resourcePercentage > 1 ) {
             throw new IllegalArgumentException(String.format("Number of resource units must be between 0 and 1, %f given",
            		 resourcePercentage));
         }
        
         resourceInformation.get(resourceTypeId).setPercentagedAmount(resourcePercentage);
	}

    @Override
    public void shutdown() {
        // not applicable
        // TODO find out expected behaviour
    }

    @Override
    public SimHost getHost() {
        return host;
    }

    @Override
    public short[] getResourceTypes() {
        short[] resourceTypes = new short[resourceInformation.size()];
        int i = 0;
        for (short type : resourceInformation.keySet()) {
            resourceTypes[i] = type;
            i++;
        }
        return resourceTypes;
    }

    @Override
    public String getName(short resourceType) {
        if (!resourceInformation.containsKey(resourceType)) {
            return null;
        }
        return resourceInformation.get(resourceType).name;
    }

    @Override
    public double getResourceCount(short resourceType) {
        checkExistence(resourceType);
        return resourceInformation.get(resourceType).currentValue;
    }

    @Override
    public double getMaxResourceCount(short resourceType) {
        checkExistence(resourceType);
        return resourceInformation.get(resourceType).maxValue;
    }

    private void checkExistence(short resourceType) {
        if (!resourceInformation.containsKey(resourceType)) {
            throw new IllegalArgumentException("Unknown resource type " + resourceType);
        }
    }

    @Override
    public void switchedState(EnergyComponent component, EnergyState oldState, EnergyState newState, long
            timeSpentInOldState) {
        Resource resource = resourceInformation.get(energyComponents.get(component));
        resource.decrease(oldState.getEnergyConsumption() *
                ((double) timeSpentInOldState / (double) Time.SECOND));
        if (resource.isVital && resource.isEmpty()) {
            // node is empty, should go offline
            for (SimNetInterface net : getHost().getNetworkComponent().getSimNetworkInterfaces()) {
                net.goOffline();
            }
            consumer.setState(ResourceConsumptionState.OFFLINE);
        }
    }

    @Override
    public boolean turnOn(EnergyComponent component) {
        return !resourceInformation.get(energyComponents.get(component)).isEmpty();
    }

    private void notifyListeners(Resource resource, double oldValue) {
        for (ResourceChangeListener listener : resourceChangeListeners) {
            listener.resourceUnitsChanged(getHost(), resource.id, oldValue, resource.currentValue, resource.isVital);
        }
    }

    private class Resource {
        private short id;
        private String name;
        private double currentValue;
        private double maxValue;
        private boolean isVital;

        void decrease(double by) {
            assert by >= 0;
            double oldValue = currentValue;
            currentValue = Math.max(0, currentValue - by);
            notifyListeners(this, oldValue);
        }

        void increase(double by) {
            assert by >= 0;
            double oldValue = currentValue;
            if (maxValue < 0) {
                currentValue += by;
            } else {
                currentValue = Math.min(maxValue, currentValue + by);
            }
            notifyListeners(this, oldValue);
        }

        // @JZ : added for the demo to set amount of resource 
        void setAmount(double amount) {
        	assert amount >= 0;        	     
        	double oldValue = currentValue;
        	currentValue = Math.min(maxValue, amount);
        	notifyListeners(this, oldValue);
        }
        
        void setPercentagedAmount(double percentage) {
        	assert percentage >= 0 && percentage <= 1;
        	double oldValue = currentValue;
        	currentValue = maxValue * percentage;
        	notifyListeners(this, oldValue);
        }
        
        boolean isEmpty() {
            return currentValue == 0;
        }
    }

	
}
