package de.tud.kom.p2psim.impl.resourcedistribution;

import de.tud.kom.p2psim.api.energy.ComponentType;
import de.tud.kom.p2psim.api.energy.EnergyComponent;
import de.tud.kom.p2psim.api.energy.EnergyEventListener;
import de.tud.kom.p2psim.api.energy.EnergyState;
import de.tud.kom.p2psim.impl.energy.DefaultEnergyState;
import de.tud.kom.p2psim.impl.util.NotImplementedException;
import de.tudarmstadt.maki.simonstrator.api.Event;
import de.tudarmstadt.maki.simonstrator.api.Time;
import de.tudarmstadt.maki.simonstrator.api.component.resourcedistribution.ResourceConsumer;
import de.tudarmstadt.maki.simonstrator.api.component.resourcedistribution.ResourceConsumptionState;
import de.tudarmstadt.maki.simonstrator.api.component.resourcedistribution.ResourceStateChangeListener;

/**
 * Implements a simple resource consumption model that periodically consumes resources based on the state of a
 * {@link ResourceConsumer}.
 */
public class StatefulResourceConsumption implements EnergyComponent, ResourceStateChangeListener {
    private final ResourceConsumer resourceConsumer;
    private final EnergyState roamingEnergyState;
    private final EnergyState headingForResourceEnergyState;
    private final EnergyState offlineEnergyState;

    private final long consumptionInterval;

    private boolean on = true;
    private EnergyState currentEnergyState;
    private long timeOfLastConsumption = 0;

    private EnergyEventListener resourceModel;

    public StatefulResourceConsumption(ResourceConsumer resourceConsumer, double roamingResourcesPerSecond, double
            headingResourcesPerSecond, long consumptionInterval) {
        this.resourceConsumer = resourceConsumer;
        this.roamingEnergyState = new DefaultEnergyState("ROAMING", roamingResourcesPerSecond);
        this.headingForResourceEnergyState = new DefaultEnergyState("HEADING_FOR_RESOURCE", headingResourcesPerSecond);
        offlineEnergyState = new DefaultEnergyState("OFFLINE_STATE", 0);

        this.consumptionInterval = consumptionInterval;
        this.currentEnergyState = mapState(resourceConsumer.getState());
        timeOfLastConsumption = Time.getCurrentTime();

        resourceConsumer.addStateChangeListener(this);

        rescheduleConsumptionTimer();
    }

    /**
     * Maps the given resource consumption state to an internally used energy state
     */
    private EnergyState mapState(ResourceConsumptionState state) {
        switch (state) {
            case OFFLINE:
                return offlineEnergyState;
            case ROAMING:
                return roamingEnergyState;
            case HEADING_FOR_RESOURCE:
                return headingForResourceEnergyState;

            default:
                throw new NotImplementedException("resource state not implemented");
        }
    }

    private void rescheduleConsumptionTimer() {
        if (!isOn()) return;
        Event.scheduleWithDelay(consumptionInterval, this, null, 0);
    }

    private void consumeResources(EnergyState oldState, EnergyState newState) {
        if (isOn()) {
            resourceModel.switchedState(this, oldState, newState, Time.getCurrentTime() - timeOfLastConsumption);
            if (!resourceModel.turnOn(this)) {
                turnOff();
            }
        }
        timeOfLastConsumption = Time.getCurrentTime();
    }

    @Override
    public void eventOccurred(Object o, int i) {
        consumeResources(currentEnergyState, currentEnergyState);
        if (isOn()) {
            rescheduleConsumptionTimer();
        }
    }

    @Override
    public void stateChanged(ResourceConsumptionState oldState, ResourceConsumptionState
            newState) {
        currentEnergyState = mapState(newState);
        if (oldState != ResourceConsumptionState.OFFLINE) { // prevent cycles
            consumeResources(mapState(oldState), mapState(newState));
        }
    }

    @Override
    public ComponentType getType() {
        return ComponentType.BASIC; // unimportant
    }

    @Override
    public void turnOff() {
        on = false;
    }

    @Override
    public boolean turnOn() {
        on = true;
        rescheduleConsumptionTimer();
        return true;
    }

    @Override
    public boolean isOn() {
        return on;
    }

    @Override
    public void setEnergyEventListener(EnergyEventListener listener) {
        this.resourceModel = listener;
    }
}
