package de.tud.kom.p2psim.impl.resourcedistribution.factory;

import de.tud.kom.p2psim.api.common.SimHost;
import de.tud.kom.p2psim.api.resourcedistribution.ResourceModel;
import de.tud.kom.p2psim.impl.resourcedistribution.DefaultResourceModel;
import de.tud.kom.p2psim.impl.resourcedistribution.StatefulResourceConsumption;
import de.tud.kom.p2psim.impl.resourcedistribution.util.RandomGaussianPartitionGenerator;
import de.tudarmstadt.maki.simonstrator.api.Host;
import de.tudarmstadt.maki.simonstrator.api.Time;
import de.tudarmstadt.maki.simonstrator.api.component.ComponentNotAvailableException;
import de.tudarmstadt.maki.simonstrator.api.component.HostComponent;
import de.tudarmstadt.maki.simonstrator.api.component.HostComponentFactory;
import de.tudarmstadt.maki.simonstrator.api.component.resourcedistribution.ResourceConsumer;

import java.util.Arrays;

/**
 * Produces a {@link DefaultResourceModel} with a single, energy-like resource using {@link StatefulResourceConsumption}
 * and a Gaussian partition of available energy amongst all hosts.
 * <p/>
 * The partition is performed using the algorithm described in {@link RandomGaussianPartitionGenerator}.
 *
 * @author Tim Feuerbach
 */
public class RandomPartitionResourceModelFactory implements HostComponentFactory {

    protected int numberOfHosts = -1;
    protected double minResourceValue = 200;
    protected double maxResourceValue = 2000;
    protected double intervalDeviationFactor = 0.1;
    protected double totalResources;
    private String totalResourcesString;

    private double[] partition = null;
    private int hostCounter = 0;

    private double consumptionInRoaming = 1;
    private double consumptionInHeading = 3;
    private long consumptionInterval = Time.SECOND;

    public static final short RESOURCE_TYPE_ID = 0;

    public static final int MAX_TRIES = 2000;

    private RandomGaussianPartitionGenerator partitionGenerator;

    public RandomPartitionResourceModelFactory() {
        partitionGenerator = new RandomGaussianPartitionGenerator();
        partitionGenerator.setMaxTries(MAX_TRIES);
    }

    @Override
    public HostComponent createComponent(Host host) {
        if (!isPartitionCreated()) {
            createPartition();
        }

        SimHost simHost = (SimHost) host;

        try {
            ResourceConsumer consumer = host.getComponent(ResourceConsumer.class);
            StatefulResourceConsumption energyComponent = new StatefulResourceConsumption(consumer,
                    consumptionInRoaming, consumptionInHeading, consumptionInterval);
            ResourceModel model = new DefaultResourceModel(simHost);
            model.registerResourceType((short) 0, "Energy", maxResourceValue, partition[hostCounter], true);
            model.registerComponent(energyComponent, (short) 0);

            hostCounter++;

            return model;
        } catch (ComponentNotAvailableException e) {
            throw new IllegalStateException("Could not create resource model: host must provide ResoureConsumer " +
                    "component.");
        }
    }

    private void createPartition() {
        if (numberOfHosts <= 0) {
            throw new IllegalArgumentException("Number of hosts not set in " + getClass().toString() + ", is: " +
                    numberOfHosts);
        }
        setTotalResources();

        assert totalResources <= maxResourceValue * numberOfHosts;

        partition = partitionGenerator.createPartition(totalResources, numberOfHosts, intervalDeviationFactor,
                minResourceValue, maxResourceValue);
    }

    private boolean isPartitionCreated() {
        return partition != null;
    }

    private void printDebugPartition() {
        System.out.println(String.format("Resource partition" +
                        " created " +
                        "for %d hosts, total resources %f, minResources %f, maxResources %f, deviation factor %f: %s",
                numberOfHosts, totalResources, minResourceValue, maxResourceValue, intervalDeviationFactor, Arrays
                        .toString(partition)));
    }

    private void setTotalResources() {
        if (totalResourcesString.endsWith("*")) {
            totalResources = Double.valueOf(totalResourcesString.substring(0, totalResourcesString.length() - 1)) *
                    numberOfHosts;
        } else {
            totalResources = Double.valueOf(totalResourcesString);
        }
    }

    public void setMinResourceValue(double minResourceValue) {
        this.minResourceValue = minResourceValue;
    }

    public void setMaxResourceValue(double maxResourceValue) {
        this.maxResourceValue = maxResourceValue;
    }

    public void setIntervalDeviationFactor(double intervalDeviationFactor) {
        this.intervalDeviationFactor = intervalDeviationFactor;
    }

    /**
     * The total resource count distributed. Must be lower than {@link #maxResourceValue} * {@link #numberOfHosts},
     * since otherwise there would be no valid partition.
     * <p/>
     * The value can either be given as a floating point absolute or, by using an asterisk at the end, as a factor
     * for the number of hosts, e.g. <code>75*</code>.
     */
    public void setTotalResources(String resources) {
        this.totalResourcesString = resources.trim();
    }

    public void setNumberOfHosts(int numberOfHosts) {
        this.numberOfHosts = numberOfHosts;
    }

    public void setConsumptionInHeading(double consumptionInHeading) {
        this.consumptionInHeading = consumptionInHeading;
    }

    public void setConsumptionInRoaming(double consumptionInRoaming) {
        this.consumptionInRoaming = consumptionInRoaming;
    }

    public void setConsumptionInterval(long consumptionInterval) {
        this.consumptionInterval = consumptionInterval;
    }
}
