package de.tud.kom.p2psim.impl.resourcedistribution.util;

import de.tud.kom.p2psim.api.util.ArrayShuffle;
import de.tudarmstadt.maki.simonstrator.api.Randoms;

import java.math.BigDecimal;
import java.util.Arrays;
import java.util.HashSet;
import java.util.Random;
import java.util.Set;

/**
 * Creates a partition of a decimal value into N intervals such that the values follow a normal distribution.
 * <p/>
 * The algorithm works as follows:
 * <ul>
 * <li>While the remaining mass to distribute are not below the maximum value, draw a normal distributed random
 * variable with the true mean (total amount to distribute / number of intervals) and a standard deviation of
 * intervalDeviationFactor * mean, truncated between minValue and maxValue (inclusive). Put the value in a bucket.</li>
 * <li>If there is mass left, distribute it among all buckets with a non-zero value, weighted by the share they got
 * in the previous step.</li>
 * <li>If one or more buckets are empty, for each of them, take a random share from the bucket with the
 * largest value and distribute it to the empty bucket, while maintaining the minValue and maxValue constraints for
 * both of them.</li>
 * </ul>
 *
 * @author Tim Feuerbach
 */
public class RandomGaussianPartitionGenerator {
    private Random random;

    private int maxTries = 2000;
    public static final BigDecimal PRECISION_TOLERANCE = new BigDecimal(1e-6);

    public RandomGaussianPartitionGenerator() {
        random = Randoms.getRandom(RandomGaussianPartitionGenerator.class);
    }

    public double[] createPartition(double totalAmount, int numberOfIntervals, double intervalDeviationFactor, double
            minValue, double maxValue) {
        return Arrays.stream(createPartition(new BigDecimal(totalAmount), numberOfIntervals, new BigDecimal
                (intervalDeviationFactor), new BigDecimal(minValue), new BigDecimal(maxValue))).mapToDouble
                (BigDecimal::doubleValue).toArray();
    }

    public BigDecimal[] createPartition(BigDecimal totalAmount, int numberOfIntervals, BigDecimal
            intervalDeviationFactor, BigDecimal
                                                minValue, BigDecimal maxValue) {
        BigDecimal[] partition = new BigDecimal[numberOfIntervals];
        BigDecimal amountLeft = totalAmount;
        BigDecimal uniformAmount = totalAmount.divide(new BigDecimal(numberOfIntervals), BigDecimal
                .ROUND_HALF_EVEN);
        BigDecimal stddev = uniformAmount.multiply(intervalDeviationFactor);

        for (int i = 0; i < numberOfIntervals; i++) {
            if (amountLeft.compareTo(maxValue) < 0) {
                partition[i] = BigDecimal.ZERO;
                continue;
            }

            partition[i] = drawTruncatedGaussian(uniformAmount, stddev,
                    minValue, amountLeft.min(maxValue));
            amountLeft = amountLeft.subtract(partition[i]);
        }

        // healing step: still resources left
        if (amountLeft.compareTo(BigDecimal.ZERO) > 0) {
            // try to distribute according to their share of the current sum
            int tries = 0;
            while (isAboveZeroWithTolerance(amountLeft)) {
                if (tries > maxTries)
                    throw new IllegalStateException(String.format("Could not heal partition: mass left to distribute:" +
                            " %f; current partition: %s", amountLeft, Arrays.toString(partition)));

                BigDecimal currentSum = Arrays.stream(partition).filter(x -> x.compareTo(maxValue) != 0).reduce(BigDecimal.ZERO, BigDecimal::add);
                BigDecimal amountLeftBeforeLoop = amountLeft;
                for (int i = 0; i < numberOfIntervals; i++) {
                	if (partition[i].compareTo(maxValue) == 0) continue;
                    BigDecimal newValue = maxValue.min(partition[i].add(partition[i].divide
                            (currentSum, BigDecimal.ROUND_HALF_EVEN).multiply(amountLeftBeforeLoop)));
                    amountLeft = amountLeft.subtract(newValue.subtract(partition[i]));
                    partition[i] = newValue;
                }

                tries++;
            }
        }

        // healing step: partitions with a value of zero (not enough resources left)
        Set<Integer> partitionsToHeal = new HashSet<>();
        for (int i = 0; i < partition.length; i++) {
            if (partition[i].compareTo(minValue) < 0) {
                partitionsToHeal.add(i);
            }
        }

        if (partitionsToHeal.size() > 0) {
            for (int partitionToHeal : partitionsToHeal) {
                int maxIdx = 0;
                for (int i = 0; i < numberOfIntervals; i++) {
                    if (partition[i].compareTo(partition[maxIdx]) > 0) {
                        maxIdx = i;
                    }
                }

                // transfer mass
                BigDecimal newMass = drawTruncatedGaussian(partition[maxIdx].divide(BigDecimal.valueOf(2), BigDecimal
                        .ROUND_HALF_EVEN), partition[maxIdx].divide(BigDecimal.valueOf(2)).multiply(
                        intervalDeviationFactor), minValue, maxValue.subtract(minValue));
                partition[partitionToHeal] = newMass;
                partition[maxIdx] = partition[maxIdx].subtract(newMass);


                if (partition[partitionToHeal].compareTo(minValue) <= 0 && partition[partitionToHeal].compareTo
                        (maxValue) >= 0) throw new IllegalStateException("Zero partition healing went wrong");
            }

        }

        assert !isAboveZeroWithTolerance(totalAmount.subtract(Arrays.stream(partition).reduce(BigDecimal.ZERO,
                BigDecimal::add).abs())) : String.format
                ("Total amount is %f but distributed amount is %f, amountLeft is %f",
                        totalAmount, Arrays.stream(partition).reduce(BigDecimal.ZERO, BigDecimal::add), amountLeft);

        ArrayShuffle.shuffle(random, partition);

        return partition;
    }

    private boolean isAboveZeroWithTolerance(BigDecimal d) {
        return d.subtract(PRECISION_TOLERANCE).compareTo(BigDecimal.ZERO) > 0;
    }

    private BigDecimal drawTruncatedGaussian(BigDecimal mean, BigDecimal stdDeviation, BigDecimal lowerBound,
                                             BigDecimal upperBound) {
        if (upperBound.compareTo(lowerBound) < 0) {
            throw new IllegalArgumentException(String.format("upperBound %f is lower than lowerBound %f", upperBound,
                    lowerBound));
        }
        for (int tries = 0; tries < maxTries; tries++) {
            // http://stackoverflow.com/questions/31754209/can-random-nextgaussian-sample-values-from-a-distribution
            // -with-different-mean
            BigDecimal randomValue = new BigDecimal(random.nextGaussian() * stdDeviation.doubleValue() + mean
                    .doubleValue());
            if (randomValue.compareTo(lowerBound) >= 0 && randomValue.compareTo(upperBound) <= 0) {
                return randomValue;
            }
        }
        throw new IllegalStateException(String.format("Could not draw truncated gaussian with mean %f, stdDeviation " +
                "%f, lowerBound %f, upperBound %f", mean, stdDeviation, lowerBound, upperBound));
    }

    public void setMaxTries(int maxTries) {
        this.maxTries = maxTries;
    }

    /**
     * For debugging purposes.
     */
    public static void main(String[] args) {
        RandomGaussianPartitionGenerator randomGaussianPartitionGenerator = new RandomGaussianPartitionGenerator();
        double[] partition;

        Integer numSamples = Integer.valueOf(args[5]);
        for (int i = 0; i < numSamples; i++) {
            partition = randomGaussianPartitionGenerator.createPartition(Double.valueOf(args[0]) * Integer.valueOf
                    (args[1]), Integer
                    .valueOf(args[1]), Double
                    .valueOf(args[2]), Double.valueOf(args[3]), Double.valueOf(args[4]));
            for (int v = 0; v < partition.length; v++) {
                System.out.print(partition[v]);
                if (v < partition.length - 1) System.out.print(";");
            }
            
            System.out.println();
        }
    }
}
